/**
 * Copyright 2011 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * 	http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.xume.easybuilders.core;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.util.LinkedHashMap;
import java.util.Map;

import org.xume.easybuilders.api.BuildStep;
import org.xume.easybuilders.api.BuildStepFactory;
import org.xume.easybuilders.api.ConstructorArgument;
import org.xume.easybuilders.api.ObjectFactory;
import org.xume.easybuilders.util.ReflectionUtils;


/**
 * @author Johan Siebens
 */
public final class ObjectBuilderInvocationHandler<T> implements InvocationHandler {

	private final Map<Method, BuildStep> buildSteps;

	private final Object[] constructorArgs;

	private ObjectFactory<T> instantiationStrategy;

	private IntrospectionResult introspectionResult;

	public ObjectBuilderInvocationHandler(ObjectFactory<T> instantiationStrategy, IntrospectionResult introspectionResult) {
		this.instantiationStrategy = instantiationStrategy;
		this.introspectionResult = introspectionResult;
		this.buildSteps = new LinkedHashMap<Method, BuildStep>();
		this.constructorArgs = new Object[introspectionResult.numberOfConstructorArguments()];
	}

	public Object invoke(Object proxy, Method method, Object[] args) {

		Object result = null;

		if (ReflectionUtils.isBuildMethod((method))) {
			result = buildObject();
		}

		if (result == null) {
			if (isConstructorArgMethod(method)) {
				ConstructorArgument constructorArgument = introspectionResult.getConstructorArgument(method);
				this.constructorArgs[constructorArgument.getIndex()] = args[0];
				result = proxy;
			}

			if (isBuildStep(method)) {
				BuildStep buildStep = buildSteps.get(method);
				if (buildStep == null) {
					BuildStepFactory<?> buildStepFactory = introspectionResult.getBuildStepFactory(method);
					buildStep = buildStepFactory.getBuildStep();
					buildSteps.put(method, buildStep);
				}
				buildStep.record(args);
				result = proxy;
			}
		}

		if (result != null) {
			return result;
		}
		else {
			throw new UnsupportedOperationException("Don't know how to execute method [" + method + "]");
		}

	}

	private Object buildObject() {
		Object result = null;
		if (instantiationStrategy != null) {
			result = instantiationStrategy.createObject();
		}
		else {
			result = new DefaultObjectFactory().createObject();
		}
		for (BuildStep buildStep : buildSteps.values()) {
			buildStep.execute(result);
		}
		return result;
	}

	private boolean isBuildStep(Method method) {
		return introspectionResult.isBuildStep(method);
	}

	private boolean isConstructorArgMethod(Method method) {
		return introspectionResult.isConstructorArg(method);
	}

	private class DefaultObjectFactory implements ObjectFactory<Object> {

		public Object createObject() {
			Class<?> targetClass = introspectionResult.getTargetClass();
			Class<?>[] types = introspectionResult.getConstructorArgumentTypes();
			return ReflectionUtils.createInstance(targetClass, types, constructorArgs);
		}

	}

}
