/**
 * Copyright 2011 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * 	http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.xume.easybuilders.util;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;

import org.xume.easybuilders.api.InstatiationException;
import org.xume.easybuilders.api.ObjectBuilder;
import org.xume.easybuilders.api.ObjectBuilderConfigurationException;


/**
 * @author Johan Siebens
 */
@SuppressWarnings("unchecked")
public final class ReflectionUtils {

	public static final Method BUILD_METHOD = buildMethod();

	public static <T> T createInstance(Class<T> targetClass) {
		try {
			Constructor<T> constructor = targetClass.getDeclaredConstructor();
			if (!constructor.isAccessible()) {
				constructor.setAccessible(true);
			}
			return constructor.newInstance();
		}
		catch (NoSuchMethodException e) {
			throw new InstatiationException("Failed to create instance of [" + targetClass.getName() + "]: no default constructor found", e);
		}
		catch (Exception e) {
			throw new InstatiationException("Failed to create instance of [" + targetClass.getName() + "]", e);
		}
	}

	public static <T> T createInstance(Class<T> targetClass, Class<?>[] parameterTypes, Object[] values) {
		try {
			Constructor<T> constructor = targetClass.getDeclaredConstructor(parameterTypes);
			if (!constructor.isAccessible()) {
				constructor.setAccessible(true);
			}
			return constructor.newInstance(values);
		}
		catch (NoSuchMethodException e) {
			throw new InstatiationException("Failed to create instance of [" + targetClass.getName() + "]: constructor not found", e);
		}
		catch (Exception e) {
			throw new InstatiationException("Failed to create instance of [" + targetClass.getName() + "]", e);
		}
	}

	public static Method findMethod(Class<?> targetClass, String methodName, Class<?>[] parameterTypes) {
		try {
			return targetClass.getMethod(methodName, parameterTypes);
		}
		catch (NoSuchMethodException e) {
			throw new ObjectBuilderConfigurationException("No match found for method [" + methodName + "] in targetClass class ["
					+ targetClass + "]", e);
		}
	}

	public static boolean isAnObjectBuilder(Type type) {
		if (type instanceof Class<?>) {
			return ObjectBuilder.class.isAssignableFrom((Class<?>) type);
		}
		else if (type instanceof ParameterizedType) {
			ParameterizedType pt = (ParameterizedType) type;
			return ObjectBuilder.class.equals(pt.getRawType());
		}
		else {
			return false;
		}
	}

	public static boolean isBuildMethod(Method method) {
		return BUILD_METHOD.equals(method);
	}

	public static <T> Class<T> targetClass(Type type) {
		if (isAnObjectBuilder(type)) {
			if (type instanceof Class<?>) {
				return doTargetClassLookup((Class<?>) ((Class<?>) type));
			}
			else if (type instanceof ParameterizedType) {
				ParameterizedType pt = (ParameterizedType) type;
				return (Class<T>) pt.getActualTypeArguments()[0];
			}
			else {
				throw new ObjectBuilderConfigurationException("Failed to locate target class for builder [" + type + "]");
			}
		}
		throw new ObjectBuilderConfigurationException("Failed to locate target class for builder [" + type + "]");
	}

	private static Method buildMethod() {
		try {
			return ObjectBuilder.class.getDeclaredMethod("build");
		}
		catch (Exception e) {
			throw new IllegalStateException(e);
		}
	}

	private static <T> Class<T> doTargetClassLookup(Class<?> clazz) {
		Type[] genericInterfaces = clazz.getGenericInterfaces();
		Class<T> result = null;
		for (Type genericInterface : genericInterfaces) {
			if (genericInterface instanceof ParameterizedType) {
				ParameterizedType pt = (ParameterizedType) genericInterface;
				if (ObjectBuilder.class.equals(pt.getRawType())) {
					result = (Class<T>) pt.getActualTypeArguments()[0];
				}
			}
			else if (genericInterface instanceof Class<?>) {
				result = doTargetClassLookup((Class<?>) genericInterface);
			}

		}
		return result;
	}

	private ReflectionUtils() {
	}

}
